// ErrContext is a way to pass friendly error messages to the user.
// Preallocates memory.

const std = @import("std");
const mem = std.mem;
const math = std.math;
const BoundedArray = std.BoundedArray;

const capacity = 1 << 16; // 64K
const ErrContext = @This();

buf: BoundedArray(u8, capacity) =
    BoundedArray(u8, capacity).init(0) catch unreachable,
overflow: bool = false,

pub fn write(self: *ErrContext, bytes: []const u8) error{Overflow}!usize {
    const can_add = capacity - self.buf.len;
    if (can_add == 0) return error.Overflow;

    self.overflow = bytes.len > can_add;
    self.buf.appendSliceAssumeCapacity(bytes[0..math.min(bytes.len, can_add)]);
    // not adding the final zero is ok, because it needs
    // to be ignored in the iterator anyway.
    _ = self.buf.append(0) catch null;

    if (self.overflow)
        return error.Overflow;

    return bytes.len;
}

const Writer = std.io.Writer(*ErrContext, error{Overflow}, write);

pub fn writer(self: *ErrContext) Writer {
    return Writer{ .context = self };
}

pub fn iterator(self: *const ErrContext) mem.SplitIterator(u8) {
    const slice = self.buf.constSlice();
    const last_byte = if (slice[slice.len - 1] == '0') slice.len - 1 else slice.len;
    return mem.split(u8, slice[0..last_byte], "\x00");
}

const testing = std.testing;

test "basics" {
    var ctx = ErrContext{};
    var wr = ctx.writer();

    try wr.writeAll("0" ** 10);
    try wr.writeAll("1" ** 10);
    try testing.expectError(error.Overflow, wr.writeAll("3" ** capacity));
    try testing.expectError(error.Overflow, wr.writeAll("foo"));

    var it = ctx.iterator();
    try testing.expectEqualSlices(u8, it.next().?, "0" ** 10);
    try testing.expectEqualSlices(u8, it.next().?, "1" ** 10);

    const long = it.next().?;
    try testing.expectEqual(@as(usize, capacity - 2 - 20), long.len);
    try testing.expectEqual(it.next(), null);
    try testing.expect(ctx.overflow);
}

test "almost overflow" {
    var ctx = ErrContext{};
    var wr = ctx.writer();

    try wr.writeAll("0" ** (capacity - 2));
    try testing.expectError(error.Overflow, wr.writeAll("11"));

    var it = ctx.iterator();
    try testing.expectEqualSlices(u8, it.next().?, "0" ** (capacity - 2));
    try testing.expectEqualSlices(u8, it.next().?, "1");
    try testing.expectEqual(it.next(), null);
}
